﻿using static Ozeki.OzHfNetworkOperation;

namespace Ozeki
{
    public static class OzHfDownloader
    {                       
        public static bool Download(OzHfDownloadParams downloadParams, out string error, IOzHfProgressReport? progress = null)
        {
            if (!IsHuggingFaceReachable())
            {
                error = "Huggingface.co is unreachable. Please check your internet connection.";
                return false;
            }
            if (!ValidateDownloadParams(downloadParams, out error)) 
            { 
                return false; 
            }
            progress ??= new OzHfProgressReport();
            var parser = new OzHfURLParser(downloadParams.DownloadUrl);
            if (!parser.IsValid)
            {
                error = "Invalid Download Url: " + downloadParams.DownloadUrl;
                return false;
            }
            if (!GetHuggingFaceRepo("https://huggingface.co/api/models/" + parser.Id, out var hf, out error, downloadParams.AccessToken)) return false;

            hf.setOutputLocation(downloadParams.OutputLocation);
            IEnumerable<OzHfDownloadManifest> manifests = hf
                .GetAllManifest()
                .AsParallel()
                .Select(x => new OzHfDownloadManifest(x.DownloadUrl, x.FileName, x.OutputPath, x.ChecksumUrl, GetContentLength(x.DownloadUrl, downloadParams.AccessToken)))
                .OrderBy(x => x.FileName);
            if(manifests.Any((x) => x.Size == -1))
            {
                error = "Can't access HuggingFace repo sizes";
                return false;
            }
            var filtered = manifests;
            if (!parser.IsRepository)
            {
                //Select files if the specified path is a folder
                filtered = manifests.Where((x) => String.Join('/', x.FileName.Split('/')[0..^1]) == parser.FilePath).ToList();
                //Select single file if path isn't a folder
                if (!filtered.Any()) filtered = manifests.Where((x) => x.FileName == parser.FilePath).ToList();
            }
            if (!filtered.Any())
            {
                error = "Specified resource does not exist: " + parser.FilePath;
                return false;
            }
            var fullDownloadSize = filtered.Sum(x => x.Size);
            var numberOfFiles = filtered.Count();
            long alreadyDownloaded = 0;
            if (Directory.Exists(downloadParams.OutputLocation + "/" + parser.Id))
            {
                var fileName = filtered.Select(x => x.OutputPath).ToList();
                foreach (var item in fileName)
                {
                    FileInfo info = new FileInfo(item);
                    if (info.Exists)
                    {
                        alreadyDownloaded += info.Length;
                    }
                }
            }

            var avalaibleSpace = new DriveInfo(Directory.GetDirectoryRoot(downloadParams.OutputLocation)).AvailableFreeSpace;
            var remainingSpace = avalaibleSpace - fullDownloadSize + alreadyDownloaded;
            if (remainingSpace < 0)
            {
                error = "Not enough space at location: " + downloadParams.OutputLocation + ". Additional space of " + OzHfDownloadDetails.FileSizeToString(-remainingSpace) + " is required.";
                return false;
            }

            if (!progress.AcceptDownloadDetails(new OzHfDownloadDetails(numberOfFiles, fullDownloadSize, avalaibleSpace, manifests, filtered, alreadyDownloaded), out var selectedManifests))
            {
                error = "Download prompt denied.";
                return false;
            }


            //Check if returned any Manifests aren't inside the repository
            if (selectedManifests.Except(manifests).Any())
            {
                error = "Selected download manifests are invalid.";
                return false;
            }

            alreadyDownloaded = 0;
            if (Directory.Exists(downloadParams.OutputLocation + "/" + parser.Id))
            {
                var fileName = selectedManifests.Select(x => x.OutputPath).ToList();
                foreach (var item in fileName)
                {
                    FileInfo info = new FileInfo(item);
                    if (info.Exists)
                    {
                        alreadyDownloaded += info.Length;
                    }
                }
            }

            fullDownloadSize = selectedManifests.Sum(x => x.Size);
            numberOfFiles = selectedManifests.Count();
            remainingSpace = avalaibleSpace - fullDownloadSize + alreadyDownloaded;
            if (remainingSpace < 0)
            {
                error = "Not enough space at location: " + downloadParams.OutputLocation + ". Additional space of " + OzHfDownloadDetails.FileSizeToString(-remainingSpace) + " is required.";
                return false;
            }

            var downloadManager = new OzHfDownloadManager(selectedManifests, progress, downloadParams);
            var success = downloadManager.Run(out error);
            return success;
        }

        private static bool ValidateDownloadParams(OzHfDownloadParams downloadParams, out string error)
        {
            error = "";
            if (downloadParams.MaxDownloadSpeedInBits <= 0)
            {
                error = "Invalid DownloadSpeed: must be an integer greater than zero";
                return false;
            }
            if (downloadParams.NumberOfRetries <= 0)
            {
                error = "Invalid NumberOfRetries: must be an integer greater than zero";
                return false;
            }
            if (downloadParams.ThreadNumber <= 0)
            {
                error = "Invalid ThreadNumber: must be an integer greater then zero";
                return false;
            }
            if (downloadParams.AccessToken != null && !CheckAccessKey(downloadParams.AccessToken, out error)) return false;
            try
            {
                Directory.CreateDirectory(downloadParams.OutputLocation);
            }
            catch (ArgumentException)
            {
                error = "Invalid Output Path: Specified path not valid: " + downloadParams.OutputLocation;
                return false;
            }
            catch (UnauthorizedAccessException)
            {
                error = "Invalid Output Path: No permission to access " + downloadParams.OutputLocation; ;
                return false;
            }
            catch (Exception ex)
            {
                error = "Invalid Output Path: " + ex.Message;
                return false;
            }
            return true;
        }
    }
}
