﻿using System.Diagnostics;
using System.Net;
using System.Net.Http.Headers;
using System.Security.Cryptography;
using System.Text;

namespace Ozeki
{
    internal class OzHfDownloadManager
    {
        IEnumerable<OzHfDownloadManifest> _manifests;
        string? _accessKey;
        IOzHfProgressReport _progressManager;
        bool _force;
        private int _numberOfThreads;
        private int _numberOfRetries;
        private OzHfThrottleShare? _throttleShare = null;
        private OzHfConnectionTester _connectionTester;

        public OzHfDownloadManager(IEnumerable<OzHfDownloadManifest> manifests, IOzHfProgressReport progressReport, OzHfDownloadParams downloadParams)
        {
            _manifests = manifests;
            _accessKey = downloadParams.AccessToken;
            _force = downloadParams.ForceDownload;
            _progressManager = progressReport;
            _numberOfThreads = downloadParams.ThreadNumber;
            _numberOfRetries = downloadParams.NumberOfRetries;
            _connectionTester = new OzHfConnectionTester(_numberOfRetries, _progressManager);
            if (downloadParams.MaxDownloadSpeedInBits != null)
            {
                _throttleShare = new OzHfThrottleShare(downloadParams.MaxDownloadSpeedInBits.Value);
            }
        }

        internal bool Run(out string error)
        {
            var options = new ParallelOptions()
            {
                MaxDegreeOfParallelism = _numberOfThreads
            };
            foreach (var item in _manifests)
            {
                _progressManager.HandleProgress(new OzHfDownloadProgress(item.FileName, OzHfDownloadProgressStatus.Wait));
            }
            try
            {
                var result = Parallel.ForEach(_manifests, options, HandleDownload);
                error = "";
                return result.IsCompleted;
            }
            catch (AggregateException ex)
            {
                if (ex.InnerExceptions.Any(x => x is OzHfConnectionTestException))
                {
                    error = "HFDownloader interrupted: Huggingface.co unreachable. Please check your internet connection.";
                }
                else
                {
                    error = "HFDownloader interrupted: Unexpected error happened while downloading: " + ex.InnerException;
                }
            }
            catch (Exception ex)
            {
                error = "HFDownloader interrupted: Unexpected error happened: " + ex.Message;
            }
            return false;
        }

       OzHfChecksumState ValidateFileChecksum(OzHfDownloadManifest manifest, string checksumpath)
        {
            using var sha256Hash = SHA256.Create();

            var fileinfo = new FileInfo(manifest.OutputPath);
            var fileSize = fileinfo.Length;

            var client = OzHfHTTPClient.Get();
            var request = new HttpRequestMessage(HttpMethod.Get, checksumpath);
            var response = client.Send(request);
            if (response.StatusCode != HttpStatusCode.OK)
            {
                return OzHfChecksumState.NO_CHECKSUM;
            }
            var rawchecksum = response.Content.ReadAsStringAsync().Result;
            if (!rawchecksum.StartsWith("version"))
            {
                return OzHfChecksumState.NO_CHECKSUM;
            }
            using var file = new FileStream(manifest.OutputPath, FileMode.Open, FileAccess.Read, FileShare.None);
            var progress = new OzHfProgressStream(file);
            var estimator = new OzHfIOPerformanceEstimator(fileSize);
            progress.UpdateProgress += (a, x) =>
            {
                var estimation = estimator.CalculateEstimation(x.BytesRead);
                _progressManager.HandleProgress(new OzHfDownloadProgress(manifest.FileName, OzHfDownloadProgressStatus.Validate, estimation));
            };
            var checksum = rawchecksum.Split("\n")[1].Substring("oid sha256:".Length);

            var bytes = sha256Hash.ComputeHash(progress);
            progress.Close();
            file.Close();
            StringBuilder builder = new StringBuilder();
            for (int i = 0; i < bytes.Length; i++)
            {
                builder.Append(bytes[i].ToString("x2"));
            }
            if (builder.ToString() == checksum)
            {
                return OzHfChecksumState.VALID;
            }
            return OzHfChecksumState.INVALID;
        }


        void HandleDownload(OzHfDownloadManifest manifest)
        {
            try
            {

                if (_force && File.Exists(manifest.OutputPath))
                {
                    File.Delete(manifest.OutputPath);
                }
                var status = CheckDownload(manifest);
                if (status == OzHfDownloadStatus.OK && !_force)
                {
                    _progressManager.HandleEvent(OzHfDownloadEvent.Success("File already downloaded: " + manifest.FileName));
                    _progressManager.HandleProgress(new(manifest.FileName, OzHfDownloadProgressStatus.OK, new OzHfIOPerformanceEstimation(100, 0, TimeSpan.Zero)));
                    return;
                }
                while (status != OzHfDownloadStatus.OK)
                {
                    try
                    {
                        if (status == OzHfDownloadStatus.FAILED)
                        {
                            status = OzHfDownloadStatus.PARTIAL;
                            Download(manifest, false);
                        }
                        else if (status == OzHfDownloadStatus.PARTIAL)
                        {
                            _progressManager.HandleEvent(OzHfDownloadEvent.Info("Resuming download: " + manifest.FileName));
                            Download(manifest, true);
                        }                     
                    }
                    catch (Exception ex)
                    {            
                        _throttleShare?.Leave(manifest.FileName);
                        if(!OzHfNetworkOperation.IsHuggingFaceReachable())
                        {
                            _progressManager.HandleEvent(OzHfDownloadEvent.Error("Network error while downloading: " + manifest.FileName));
                            _connectionTester.TestConnection();
                        }
                        else
                        {
                            _progressManager.HandleEvent(OzHfDownloadEvent.Error("Error while downloading  " + manifest.FileName + ": " + ex.Message));
                        }
                    }
                    try
                    {
                        status = CheckDownload(manifest);
                        if (status == OzHfDownloadStatus.FAILED)
                        {
                            _progressManager.HandleEvent(OzHfDownloadEvent.Error("Download failed, restarting: " + manifest.FileName));
                            File.Delete(manifest.OutputPath);
                        }
                    }
                    catch (Exception ex)
                    {
                        _progressManager.HandleEvent(OzHfDownloadEvent.Error("Error while checking " + manifest.FileName + ": " + ex.Message));
                        status = OzHfDownloadStatus.FAILED;
                    }
                }
            }
            catch (Exception ex)
            {
                _progressManager.HandleEvent(OzHfDownloadEvent.Error(ex.Message));
                _throttleShare?.Leave(manifest.FileName);
                throw;
            }

            _progressManager.HandleEvent(OzHfDownloadEvent.Success(manifest.FileName + " finished."));
            _progressManager.HandleProgress(new OzHfDownloadProgress(manifest.FileName, OzHfDownloadProgressStatus.OK));
            _throttleShare?.Leave(manifest.FileName);
        }

        OzHfDownloadStatus CheckDownload(OzHfDownloadManifest manifest)
        {
            var fileinfo = new FileInfo(manifest.OutputPath);
            if (fileinfo.Directory == null)
            {
                return OzHfDownloadStatus.FAILED;
            }
            if (!fileinfo.Directory.Exists) fileinfo.Directory.Create();
            if (!fileinfo.Exists) return OzHfDownloadStatus.FAILED;

            var contentLength = OzHfNetworkOperation.GetContentLength(manifest.DownloadUrl);
            var downloadedLength = fileinfo.Length;

            if (downloadedLength < contentLength) return OzHfDownloadStatus.PARTIAL;
            if (downloadedLength > contentLength) return OzHfDownloadStatus.FAILED;

            var check = ValidateFileChecksum(manifest, manifest.ChecksumUrl);

            if ((check == OzHfChecksumState.VALID || check == OzHfChecksumState.NO_CHECKSUM)) return OzHfDownloadStatus.OK;
            return OzHfDownloadStatus.FAILED;
        }

        void Download(OzHfDownloadManifest manifest, bool partial = false)
        {
            string url = manifest.DownloadUrl;
            string outputPath = manifest.OutputPath;

            const int BUFFER_SIZE = 32 * 1024;
            HttpClient client = OzHfHTTPClient.Get();
            var request = new HttpRequestMessage(HttpMethod.Get, url);
            request.RequestUri = new Uri(url);
            if (_accessKey != null)
            {
                request.Headers.Authorization = new AuthenticationHeaderValue("Bearer", _accessKey);
            }
            long start = 0;
            if (partial)
            {
                var fileInfo = new FileInfo(outputPath);
                var resumeByte = fileInfo.Length;
                start = resumeByte;
                request.Headers.Range = new RangeHeaderValue(resumeByte - 0, null);
            }
            var response = new HttpResponseMessage();
            try
            {
                response = client.SendAsync(request, HttpCompletionOption.ResponseHeadersRead).Result;
            }
            catch (Exception ex)
            {
                _progressManager.HandleEvent(OzHfDownloadEvent.Error(ex.Message));
                return;
            }
            if (!response.IsSuccessStatusCode)
            {
                var msg = response.Content.ReadAsStringAsync().Result;
                Console.WriteLine(response.StatusCode);
                return;
            }
            using var responseStream = response.Content.ReadAsStreamAsync().Result;
            using var throttleStream = new OzHfThrottledStream(responseStream, null);
            _throttleShare?.Register((x) => throttleStream.MaxBitsPerSecond = x, manifest.FileName);
            var buffer = new byte[BUFFER_SIZE];
            var fileMode = partial ? FileMode.Append : FileMode.Create;
            using var outputFileStream = new BufferedStream(new FileStream(outputPath, fileMode, FileAccess.Write, FileShare.Read), BUFFER_SIZE);
            int bytesRead = 0;
            var estimator = new OzHfIOPerformanceEstimator(manifest.Size, start);
            bool success = false;
            Stopwatch sw = new Stopwatch();
            sw.Start();
            string error = "";
            Thread read = new Thread(() =>
            {
                try
                {
                    do
                    {
                        sw.Restart();
                        bytesRead = throttleStream.Read(buffer, 0, BUFFER_SIZE);
                        var estimation = estimator.CalculateEstimation(bytesRead);
                        outputFileStream.Write(buffer, 0, bytesRead);
                        var report = new OzHfDownloadProgress(manifest.FileName, OzHfDownloadProgressStatus.Download, estimation);
                        _progressManager.HandleProgress(report);
                        success = true;
                    } while (bytesRead > 0);
                }
                catch (Exception ex)
                {
                    success = false;
                    error = ex.Message;
                    return;
                }
            });

            read.Start();

            while (read.IsAlive)
            {
                Thread.Sleep(100);
                if (sw.ElapsedMilliseconds > 10000)

                {
                    read.Interrupt();
                    _throttleShare?.Leave(manifest.FileName);
                    throw new Exception("Download process doesn't responded for too long.");
                }
            }
            _throttleShare?.Leave(manifest.FileName);
            read.Join();
            if (!success) throw new Exception("Download process wasn't successful: " + error);
        }
    }
}


